/* 
 * Copyright (C) 2006 Johannes Hausensteiner
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
 
/* 
 * Filename: hex2mem.c
 *
 * Function: convert Intel HEX file into ispLEVER MEM file format
 * 
 *
 * Changelog
 *
 *  1.0  25.Jul.2006   JH   new
 *
 */


#define VER "1.0"

#ifdef WIN32
#include <windows.h>
#endif

#ifdef __MSDOS__
#include <dir.h>             /* MAXPATH */
#endif

#ifdef __linux__
#include <limits.h>          /* PATH_MAX */
#endif

#include <stdio.h>
#include <string.h>


/******************************  Constants  *******************************/
#define BUFLEN 200

#ifdef WIN32
#define PROGNAME "hex2mem.exe"
#elif defined __linux__
#define PROGNAME "hex2mem"
#define MAX_PATH PATH_MAX
#elif defined __MSDOS__
#define PROGNAME "HEX2MEM.EXE"
#define MAX_PATH MAXPATH
#else
#error "Unsupported platform!"
#endif


/*******************************  Macros  *********************************/
#ifndef HIBYTE
#define HIBYTE(x) (((x)>>8)&0xff)
#endif
#ifndef LOBYTE
#define LOBYTE(x) ((x)&0xff)
#endif


/*******************************  Globals  ********************************/
enum REC_TYPE
{
  REC_DATA,
  REC_END
};


/**************************  Function Prototypes  *************************/
int atohex (char *s);             /* convert string to HEX number         */


/* main */
/**************************************************************************/
int main (int argc, char *argv[])
/***************************************************************************
* globals  :                                                               *
* called by:                                                               *
* calls    : atohex()                                                      *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* 0 .. OK                                                                  *
***************************************************************************/
{
  FILE *in, *out;
  char inname[MAX_PATH], outname[MAX_PATH];
  char *p;
  unsigned int linecnt, bytecount, address, type, i, j;
  unsigned char data[BUFLEN/2], sum, checksum;
  int is_filter;
  char buf[BUFLEN], tmp[12];

  is_filter = 0;
  if (argc == 1)
  {
    /* run as filter */
    in = stdin;
    out = stdout;
    is_filter = 1;
  } else
  {
    in = NULL;
    out = NULL;
    for (i=1; i<argc; i++)
    {
      if (*argv[i] == '-')
      {
        argv[i] ++;
        if (*argv[i] == '-')
          argv[i] ++;
        switch (*argv[i])
        {
          case 'v':
            printf ("\n%s V%s, %s\n\n", PROGNAME, VER, __DATE__);
            return 0;
          case '\0':
            if (in == NULL)
            {
              in = stdin;
              is_filter = 1;
              break;
            } else if (out == NULL)
            {
              out = stdout;
              is_filter = 1;
              break;
            }
            /* fall through */
          case 'h':
          default:
            fprintf (stderr,
                     "\nusage: %s infile[.hex] [outfile[.mem]]\n",PROGNAME);
            return 0;
        } /* switch() */
      } /* is switch character */
      if (in == NULL)
      {
        strcpy (inname, argv[i]);
        p = strrchr (inname, '.');
        if (p == NULL)
          strcat (inname, ".hex");
        in = fopen (inname, "rt");
        if (in == NULL)
        {
          perror (inname);
          return 1;
        }
        continue;
      } else if (out == NULL)
      {
        strcpy (outname, argv[i]);
        if (outname[0] == '\0')
          strcpy (outname, "$$file$$");
        p = strrchr (outname, '.');
        if (p == NULL)
          strcat (outname, ".mem");
        out = fopen (outname, "wt");
        if (out == NULL)
        {
          perror (outname);
          return 1;
        }
      }
    } /* for loop */
  }
  if (!is_filter)
    printf ("\n%s V%s, %s\n", PROGNAME, VER, __DATE__);
  if (out == NULL)
  {
    strcpy (outname, inname);
    p = strrchr (outname, '.');
    if (p)
      *p = '\0';
    strcat (outname, ".mem");
    out = fopen (outname, "wt");
    if (out == NULL)
    {
      fclose (in);
      perror (outname);
      return 1;
    }
  }
  
  /* process the input file */
  if (!is_filter)
    printf ("\nProcessing file \"%s\" ...    ", inname);

  linecnt = 0;
  while (fgets (buf, BUFLEN, in))
  {
    linecnt ++;
    if (!is_filter)
      printf ("\b\b\b\b%4d", linecnt);

    /* read the .hex file */
    tmp[0] = buf[1];
    tmp[1] = buf[2];
    tmp[2] = '\0';
    bytecount = atohex (tmp);

    tmp[0] = buf[3];
    tmp[1] = buf[4];
    tmp[2] = buf[5];
    tmp[3] = buf[6];
    tmp[4] = '\0';
    address = atohex (tmp);

    tmp[0] = buf[7];
    tmp[1] = buf[8];
    tmp[2] = '\0';
    type = atohex (tmp);

    sum = bytecount + HIBYTE(address) + LOBYTE(address) + type;
    for (i=0; i<=bytecount; i++)
    {
      j = i * 2 + 9;
      tmp[0] = buf[j];
      j ++;
      tmp[1] = buf[j];
      tmp[2] = '\0';
      if (i < bytecount)
      {
        data[i] = atohex (tmp);
        sum += data[i];
      } else
      {
        checksum = atohex (tmp);
        sum = (-sum) & 0xff;
      }
    }
    if (sum != checksum)
    {
      fprintf(stderr,"\nError in line %d: checksum failure (%2.2x/%2.2x)\n",
        linecnt, (unsigned char)sum, (unsigned char)checksum);
    }
    /* write the .mem file */
    if (type != REC_END)
    {
      fprintf (out, "%4.4X:", address);
      for (i=0; i<bytecount; i++)
      {
        fprintf (out, " %2.2X", data[i]);
        if (i == 7)
          fputc (' ', out);
      }
      fputc ('\n', out);
    }
  }  
  if (!is_filter)
  {
    printf ("\b\b\b\b%4d lines done.\n", linecnt);
    fclose (in);
    fclose (out);
  }
  return 0;
}

/* atohex */
/**************************************************************************/
int atohex (char *s)              /* convert string to HEX number         */
/***************************************************************************
* globals     :                                                            *
* called from : main()                                                     *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Convert string to HEX number                                             *
*                                                                          *
* Returns                                                                  *
* --------                                                                 *
* number (as integer)                                                      *
***************************************************************************/
{
   int i, digit, number;

   i = 0;
   number = 0;
   while (s[i]) {
     switch (toupper(s[i])) {
       case '0': digit =  0; break;
       case '1': digit =  1; break;
       case '2': digit =  2; break;
       case '3': digit =  3; break;
       case '4': digit =  4; break;
       case '5': digit =  5; break;
       case '6': digit =  6; break;
       case '7': digit =  7; break;
       case '8': digit =  8; break;
       case '9': digit =  9; break;
       case 'A': digit = 10; break;
       case 'B': digit = 11; break;
       case 'C': digit = 12; break;
       case 'D': digit = 13; break;
       case 'E': digit = 14; break;
       case 'F': digit = 15; break;
     }
     number = (number << 4) + digit;
     i++;
   }
/* printf("Number = %i, 0x%X\n", number, number);*/
   return (number);
}
